@@ -975,6 +975,7 @@ frobpoltrunc(GEN hgm, GEN t, long c, long p, long B, long *pF)
   m = (mi == D && c == C_TAME1 && !odd(DEG))? mi: mi+1;
   s = cgetg(m+1, t_POL); s[1] = evalsigne(1)|evalvarn(0);
   for (f = 1; f < m; f++) gel(s, f+1) = negi( hgmtrace(hgm, p, f, t, c) );
+  s = RgX_renormalize_lg(s, m+1);
   *pF = 0; s = RgXn_expint(s, m);
   if (mi == D) return s;
   if (c == C_TAME1)
@@ -1260,7 +1261,7 @@ hgmU(GEN hgm, long p, long f, GEN t, long dfp)
   long q = vp[f+1], i;
   GEN Q = cgetg(q+1, t_POL);
   Q[1] = evalsigne(1)|evalvarn(0);
-  for (i = 2; i <= q; i++) gel(Q, i) = hgmQ(hgm, p, f, vp, i, ZP);
+  for (i = 2; i <= q; i++) gel(Q, i) = hgmQ(hgm, p, f, vp, i-2, ZP);
   t = p == 2? gen_1: gmul(hgm_get_MVALUE(hgm), t);
   return gerepileupto(av, hgmH(Q, p, f, dfp, t));
 }
@@ -2031,7 +2032,7 @@ lfunhgmwild(GEN L, GEN H, GEN t, GEN BAD, long pole, GEN hint, long bitprec)
     gel(L, 5) = gel(vN, lN - 1);
     if (DEBUGLEVEL) err_printf("\nmaxN = %ld\n", itos(gel(L,5)));
     lim = lfunthetacost(L, t0i, 0, bitprec);
-    vroots = dirpowers(lim, sstoQ(2, d), prec + EXTRAPRECWORD);
+    vroots = dirpowers(lim, sstoQ(2, d), prec + EXTRAPREC64);
     an0 = hgmcoefs(H, t, lim);
     if (pole)
     {
