@@ -98,14 +98,14 @@ static GEN
 qfbmat(GEN q)
 {
   GEN a = gel(q,1), b = shifti(gel(q,2), -1), c = gel(q,3);
-  return mkmat2(mkcol2(a, b), mkcol2(b, c));
+  retmkmat22(a, b, b, c);
 }
 /* 2*qfbmat(q) */
 static GEN
 qfbmat2(GEN q)
 {
   GEN a = shifti(gel(q,1), 1), b = gel(q,2), c = shifti(gel(q,3), 1);
-  return mkmat2(mkcol2(a, b), mkcol2(b, c));
+  retmkmat22(a, b, b, c);
 }
 /* Given a symmetric matrix G over Z, compute the Witt invariant of G at p
  * v = det_minors(G) [G diagonalized]; assume that none of the v[i] is 0. */
@@ -180,7 +180,7 @@ qflllgram_indef(GEN G, long base, int *fail)
   /* G = M~*D*M, D diagonal, DM=|D|*M, g =  M~*|D|*M */
   g = ZM_transmultosym(M,DM);
   S = lllgramint(Q_primpart(g));
-  R = qftriv(qf_apply_ZM(G,S), NULL, base);
+  R = qftriv(qf_ZM_apply(G,S), NULL, base);
   switch(typ(R))
   {
     case t_COL: return ZM_ZC_mul(S,R);
@@ -219,7 +219,7 @@ qflllgram_indefgoon(GEN G)
   U1 = gel(red,2);
   G2 = gel(red,1); /* G2[1,1] = 0 */
   U2 = gel(ZV_extgcd(row(G2,1)), 2);
-  G3 = qf_apply_ZM(G2,U2);
+  G3 = qf_ZM_apply(G2,U2);
   U = ZM_mul(U1,U2); /* qf_apply(G,U) = G3 */
   /* G3[1,] = [0,...,0,g], g^2 | det G */
   g = gcoeff(G3,1,n);
@@ -228,8 +228,7 @@ qflllgram_indefgoon(GEN G)
   /* G3[n,n] reduced mod 2g */
   if (n == 2) return mkvec2(G3,U);
   V = rowpermute(vecslice(G3, 2,n-1), mkvecsmall2(1,n));
-  A = mkmat2(mkcol2(gcoeff(G3,1,1),gcoeff(G3,1,n)),
-             mkcol2(gcoeff(G3,1,n),gcoeff(G3,2,2)));
+  A = mkmat22(gcoeff(G3,1,1),gcoeff(G3,1,n),gcoeff(G3,1,n),gcoeff(G3,2,2));
   B = ground(RgM_neg(QM_mul(QM_inv(A), V)));
   U3 = matid(n);
   for (j = 2; j < n; j++)
@@ -237,14 +236,14 @@ qflllgram_indefgoon(GEN G)
     gcoeff(U3,1,j) = gcoeff(B,1,j-1);
     gcoeff(U3,n,j) = gcoeff(B,2,j-1);
   }
-  G4 = qf_apply_ZM(G3,U3); /* the last column of G4 is reduced */
+  G4 = qf_ZM_apply(G3,U3); /* the last column of G4 is reduced */
   U = ZM_mul(U,U3);
   if (n == 3) return mkvec2(G4,U);
 
   red = qflllgram_indefgoon(matslice(G4,2,n-1,2,n-1));
   if (typ(red) == t_MAT) return mkvec2(G4,U);
   /* Let U5:=matconcat(diagonal[1,red[2],1])
-   * return [qf_apply_ZM(G5, U5), U*U5] */
+   * return [qf_ZM_apply(G5, U5), U*U5] */
   G5 = gel(red,1);
   U5 = gel(red,2);
   G6 = cgetg(n+1,t_MAT);
@@ -261,7 +260,7 @@ qflllgram_indefgoon(GEN G)
   return mkvec2(G6, shallowconcat1(U6));
 }
 
-/* qf_apply_ZM(G,H),  where H = matrix of \tau_{i,j}, i != j */
+/* qf_ZM_apply(G,H),  where H = matrix of \tau_{i,j}, i != j */
 static GEN
 qf_apply_tau(GEN G, long i, long j)
 {
@@ -293,10 +292,10 @@ qflllgram_indefgoon2(GEN G)
     b = diviiexact(b,d);
   }
   /* for U2 = [-u,-b,0;-v,a,0;0,0,1]
-   * G3 = qf_apply_ZM(G2,U2) has known last row (-d, 0, 0),
+   * G3 = qf_ZM_apply(G2,U2) has known last row (-d, 0, 0),
    * so apply to principal_minor(G3,2), instead */
-  U2 = mkmat2(mkcol2(negi(u),negi(v)), mkcol2(negi(b),a));
-  G3 = qf_apply_ZM(principal_minor(G2,2),U2);
+  U2 = mkmat22(negi(u),negi(b),negi(v),a);
+  G3 = qf_ZM_apply(principal_minor(G2,2),U2);
   r3 = gel(r,3);
   r = ZV_ZM_mul(mkvec2(gel(r,1),gel(r,2)),U2);
 
@@ -402,9 +401,10 @@ prod_fm(GEN f, long i)
  * is integral and has minimal determinant.
  * In dimension 3 or 4, may return a prime p if the reduction at p is
  * impossible because of local nonsolvability.
- * P,E = factor(+/- det(G)), "prime" -1 is ignored. Destroy E. */
+ * P,E = factor(+/- det(G)), d = det(G) "prime" -1 is ignored,
+ * Either E or d should be NULL, but not both */
 static GEN
-qfminimize_fact(GEN G, GEN P, GEN E, long loc)
+qfminimize_fact(GEN G, GEN P, GEN E, GEN d, long loc)
 {
   GEN U = NULL, Ker = NULL, faE, faP;
   long n = lg(G)-1, lP = lg(P), i;
@@ -412,8 +412,10 @@ qfminimize_fact(GEN G, GEN P, GEN E, long loc)
   faE = vecsmalltrunc_init(lP);
   for (i = 1; i < lP; i++)
   {
+    long Ei, vp, wp;
     GEN p = gel(P,i);
-    long vp = E[i], wp = vp;
+    if (is_pm1(p)) continue;
+    Ei = E ? E[i]: Z_pval(d, p); vp = Ei; wp = vp;
     if (!vp) continue;
     if (DEBUGLEVEL >= 3) err_printf("qfminimize: for %Ps^%ld:", p,vp);
     while (vp)
@@ -424,13 +426,13 @@ qfminimize_fact(GEN G, GEN P, GEN E, long loc)
       while (vp) /* loop until vp <= n */
       {
         if (DEBUGLEVEL>=3 && vp <= wp)
-        { err_printf(" %ld%%", (E[i]-vp)*100/E[i]); wp -= E[i]/100; }
+        { err_printf(" %ld%%", (Ei-vp)*100/Ei); wp -= Ei/100; }
         /* The case vp = 1 can be minimized only if n is odd. */
         if (vp == 1 && !odd(n)) break;
         Ker = kermodp(G,p, &dimKer); /* dimKer <= vp */
         if (DEBUGLEVEL >= 4) err_printf("    dimKer = %ld\n",dimKer);
         if (dimKer == n) break;
-        G = qf_apply_ZM(G, Ker);
+        G = qf_ZM_apply(G, Ker);
         /* 1st case: dimKer < vp */
         /* then the kernel mod p contains a kernel mod p^2 */
         if (dimKer >= vp) break;
@@ -452,10 +454,10 @@ qfminimize_fact(GEN G, GEN P, GEN E, long loc)
           blocks4(G, dimKer,n, &A,&B,&C); A = ZsymM_Z_divexact(A, p);
           K2 = kermodp(A, p, &dimKer2);
           /* Write G = [pA,B;B~,C] and apply [K2/p,0;0,Id] by blocks */
-          A = qf_apply_ZM(A,K2); ZsymM_Z_divexact_partial(A, dimKer2, p);
+          A = qf_ZM_apply(A,K2); ZsymM_Z_divexact_partial(A, dimKer2, p);
           B = ZM_transmul(B,K2);
           for (j = 1; j <= dimKer2; j++) gel(B,j) = ZC_Z_divexact(gel(B,j), p);
-          G = shallowmatconcat(mkmat2(mkcol2(A,B), mkcol2(shallowtrans(B), C)));
+          G = shallowmatconcat(mkmat22(A,shallowtrans(B),B,C));
           /* Ker *= [K2,0;0,Id] */
           B = ZM_mul(vecslice(Ker,1,dimKer),K2);
           for (j = 1; j <= dimKer2; j++) gel(B,j) = RgC_Rg_div(gel(B,j), p);
@@ -516,7 +518,7 @@ qfminimize_fact(GEN G, GEN P, GEN E, long loc)
         sol = Q_primpart(sol);
         if (DEBUGLEVEL >= 4) err_printf("    sol = %Ps\n", sol);
         Ker = completebasis(vecextend(sol,n), 1);
-        G = qf_apply_ZM(G, Ker);
+        G = qf_ZM_apply(G, Ker);
         for (j = 1; j < n; j++)
           gcoeff(G,n,j) = gcoeff(G,j,n) = diviiexact(gcoeff(G,j,n), p);
         gcoeff(G,n,n) = diviiexact(gcoeff(G,n,n), sqri(p));
@@ -553,8 +555,8 @@ qfminimize_fact(GEN G, GEN P, GEN E, long loc)
   if (!U) U = matid(n);
   else
   { /* apply LLL to avoid coefficient explosion */
-    GEN u = lllint(Q_primpart(U));
-    G = qf_apply_ZM(G, u);
+    GEN u = ZM_lll(Q_primpart(U), .99, LLL_IM);
+    G = qf_ZM_apply(G, u);
     U = QM_mul(U, u);
   }
   return mkvec4(G, U, faP, faE);
@@ -570,22 +572,36 @@ check_symmetric(GEN G)
       if (!equalii(gcoeff(G,i,j), gcoeff(G,j,i)))
         pari_err_TYPE("qfsolve [not symmetric]",G);
 }
+/* assume G symmetric and integral */
+static void
+symmetric_non0_coeff(GEN G, long *pi, long *pj)
+{
+  long i, j, l = lg(G);
+  *pi = *pj = 0;
+  for (i = 1; i < l; i++)
+    for (j = 1; j <= i; j++)
+      if (signe(gcoeff(G,i,j))) { *pi = i; *pj = j; return; }
+}
 
 GEN
 qfminimize(GEN G)
 {
   pari_sp av = avma;
-  GEN d, F, H;
-  long n = lg(G)-1;
+  GEN c, d, F, H, U;
+  long i, j, n = lg(G)-1;
   if (typ(G) != t_MAT) pari_err_TYPE("qfminimize", G);
   if (n == 0) pari_err_DOMAIN("qfminimize", "dimension" , "=", gen_0, G);
   if (n != nbrows(G)) pari_err_DIM("qfminimize");
   G = Q_primpart(G); RgM_check_ZM(G, "qfminimize");
   check_symmetric(G);
   d = ZM_det(G);
+  if (!signe(d)) pari_err_DOMAIN("qfminimize", "det" , "=", gen_0, gen_0);
   F = absZ_factor(d);
-  H = qfminimize_fact(G, gel(F,1),  ZV_to_zv(gel(F,2)), 0);
-  return gerepilecopy(av, mkvec2(gel(H,1), gel(H,2)));
+  H = qfminimize_fact(G, gel(F,1), ZV_to_zv(gel(F,2)), NULL, 0);
+  symmetric_non0_coeff(G, &i, &j);
+  U = gel(H,2); H = gel(H,1);
+  c = gdiv(gcoeff(H,i,j), RgV_dotproduct(gel(U,i), RgM_RgC_mul(G, gel(U,j))));
+  return gerepilecopy(av, mkvec3(H, U, c));
 }
 
 /* CLASS GROUP COMPUTATIONS */
@@ -645,7 +661,7 @@ qfb_factorback(GEN gen, GEN e, GEN isqrtD)
 /* unit form discriminant 4d */
 static GEN
 id(GEN d)
-{ return mkmat2(mkcol2(gen_1,gen_0),mkcol2(gen_0,negi(d))); }
+{ retmkmat22(gen_1, gen_0, gen_0, negi(d)); }
 
 /* Shanks/Bosma-Stevenhagen algorithm to compute the 2-Sylow of the class
  * group of discriminant D. Only works for D = fundamental discriminant.
@@ -762,7 +778,7 @@ qftriv(GEN G, GEN R, long base)
       s = col_ei(n,i); gel(s,i-1) = gen_m1;
       if (!base) return s;
       H = matid(n); gel(H,i) = gel(H,1); gel(H,1) = s;
-      return mkvec2(qf_apply_ZM(G,H),H);
+      return mkvec2(qf_ZM_apply(G,H),H);
     }
   if (!R) return G; /* fail */
   /* case 3: a principal minor is 0 */
@@ -771,7 +787,7 @@ qftriv(GEN G, GEN R, long base)
   if (!base) return s;
   H = completebasis(s, 0);
   gel(H,n) = ZC_neg(gel(H,1)); gel(H,1) = s;
-  return mkvec2(qf_apply_ZM(G,H),H);
+  return mkvec2(qf_ZM_apply(G,H),H);
 }
 
 /* p a prime number, G 3x3 symmetric. Finds X!=0 such that X^t G X = 0 mod p.
@@ -827,11 +843,23 @@ qfsolvemodp(GEN G, GEN p)
 static  GEN
 qfsolve_i(GEN G)
 {
-  GEN M, signG, Min, U, G1, M1, G2, M2, solG2, P, E;
-  GEN solG1, sol, Q, d, dQ, detG2, fam2detG;
+  GEN M, signG, Min, U, G1, M1, G2, M2, solG2, P = NULL, E = NULL;
+  GEN solG1, sol, Q, d, dQ, detG2;
   long n, np, codim, dim;
   int fail;
 
+  if (typ(G) == t_VEC && lg(G)==3)
+  {
+    P = gel(G,2);
+    G = gel(G,1);
+    if (typ(P)==t_MAT)
+    {
+      if (!is_Z_factornon0(P)) pari_err_TYPE("qfsolve", P);
+      P = gel(P,1);
+    } else
+      if (!is_vec_t(typ(P)) || !RgV_is_ZVnon0(P))
+        pari_err_TYPE("qfsolve", P);
+  }
   if (typ(G) != t_MAT) pari_err_TYPE("qfsolve", G);
   n = lg(G)-1;
   if (n == 0) pari_err_DOMAIN("qfsolve", "dimension" , "=", gen_0, G);
@@ -875,13 +903,14 @@ qfsolve_i(GEN G)
   }
 
   /* factorization of the determinant */
-  fam2detG = absZ_factor(d);
-  P = gel(fam2detG,1);
-  E = ZV_to_zv(gel(fam2detG,2));
-  /* P,E = factor(|det(G)|) */
+  if (!P)
+  {
+    GEN F = absZ_factor(d);
+    P = gel(F,1); E = ZV_to_zv(gel(F,2));
+  }
 
   /* Minimization and local solubility */
-  Min = qfminimize_fact(G, P, E, 1);
+  Min = qfminimize_fact(G, P, E, d, 1);
   if (typ(Min) == t_INT) return Min;
 
   M = QM_mul(M, gel(Min,2));
@@ -931,7 +960,7 @@ qfsolve_i(GEN G)
       signG[1]++;
     G1 = shallowmatconcat(diagonal_shallow(mkvec2(G,aux)));
     /* P,E = factor(|det G1|) */
-    Min = qfminimize_fact(G1, P, E, 1);
+    Min = qfminimize_fact(G1, P, E, NULL, 1);
     G1 = gel(Min,1);
     M1 = gel(Min,2);
     P = gel(Min,3);
@@ -991,9 +1020,8 @@ qfsolve_i(GEN G)
     G2 = shallowmatconcat(diagonal_shallow(mkvec2(G1,ZM_neg(Q))));
     /* Minimization of G2 */
     detG2 = mulii(d, ZM_det(Q));
-    for (i = 1; i < lfactdP; i++) factdE[i] = Z_pval(detG2, gel(factdP,i));
     /* factdP,factdE = factor(|det G2|) */
-    Min = qfminimize_fact(G2, factdP,factdE,1);
+    Min = qfminimize_fact(G2, factdP, NULL, detG2, 1);
     M2 = gel(Min,2);
     G2 = gel(Min,1);
   }
@@ -1052,7 +1080,7 @@ qfparam(GEN G, GEN sol, long fl)
   sol = Q_primpart(sol); RgV_check_ZV(sol,"qfsolve");
   /* build U such that U[,3] = sol, and |det(U)| = 1 */
   U = completebasis(sol,1);
-  G1 = qf_apply_ZM(G,U); /* G1 has a 0 at the bottom right corner */
+  G1 = qf_ZM_apply(G,U); /* G1 has a 0 at the bottom right corner */
   a = shifti(gcoeff(G1,1,2),1);
   b = shifti(negi(gcoeff(G1,1,3)),1);
   c = shifti(negi(gcoeff(G1,2,3)),1);
@@ -1067,7 +1095,7 @@ qfparam(GEN G, GEN sol, long fl)
     a = gel(v,1);
     b = gmul2n(gel(v,2),-1);
     c = gel(v,3);
-    U = qflllgram_indef(mkmat2(mkcol2(a,b),mkcol2(b,c)), 1, &fail);
+    U = qflllgram_indef(mkmat22(a, b, b, c), 1, &fail);
     U = gel(U,2);
     a = gcoeff(U,1,1); b = gcoeff(U,1,2);
     c = gcoeff(U,2,1); d = gcoeff(U,2,2);
