@@ -582,7 +582,7 @@ u_forprime_sieve_arith_init(forprime_t *T, struct pari_sieve *psieve,
 #else
   const ulong UPRIME_MAX = 4294967291UL;
 #endif
-  ulong P, P2, Y, sieveb;
+  ulong Plim, P, P2, Y, sieveb;
 
   if (!odd(b) && b > 2) b--;
   if (a > b || b < 2)
@@ -624,25 +624,28 @@ u_forprime_sieve_arith_init(forprime_t *T, struct pari_sieve *psieve,
   }
   else
     u_forprime_set_prime_table(T, a);
+  if (T->strategy == PRST_none) T->strategy = PRST_unextprime;
+  /* now strategy is either PRST_diffptr or PRST_unextprime */
 
   P2 = (P & HIGHMASK)? 0 : P*P;
   sieveb = b; if (P2 && P2 < b) sieveb = P2;
   /* maxprime^2 >= sieveb */
-  Y = sieveb - maxprimelim(); /* length of sievable interval */
+  Plim = maxprimelim();
+  if (a <= Plim) a = Plim + 1;
+  if (sieveb < a + 16) return 1;
+  Y = sieveb - a + 1; /* number of integers in sievable interval > 16 */
   P = usqrt(sieveb); /* largest sieving prime */
   /* FIXME: should sieve as well if q != 1, adapt sieve code */
-  if (q != 1 || (P2 && P2 <= a) || 3/M_LN2 * Y < uprimepi(P))
-  /* Sieve unimplemented/impossible/too costly. Cost model is
+  if (q == 1 && (!P2 || P2 > a) && 3/M_LN2 * Y >= uprimepi(P))
+  /* Sieve implemented & possible & not too costly. Cost model is
    * - nextprime: about Y / log(b) primes to test [neglect cost for composites]
    *   individual cost average = 3 log2(b) mulmod, total = 3 Y / log(2) mulmod
    * - sieve: pi(P) mod + Y loglog(b) add
    * Since loglog(b) < 4, and add < 10*mulmod, we neglect the Y loglog(b) term.
    * We have mod < mulmod < 2*mod; for now, assume mulmod ~ mod. */
-  { if (T->strategy==PRST_none) T->strategy = PRST_unextprime; }
-  else
-  { /* worth sieving */
-    if (T->strategy==PRST_none) T->strategy = PRST_sieve;
-    sieve_init(T, maxuu(maxprimelim()+1, a), sieveb);
+  {
+    if (T->strategy == PRST_unextprime) T->strategy = PRST_sieve;
+    sieve_init(T, a, sieveb);
   }
   return 1;
 }
@@ -861,7 +864,7 @@ u_forprime_next(forprime_t *T)
     }
   }
   if (T->strategy == PRST_sieve)
-  {
+  { /* require sieveb - a >= 16 */
     ulong n;
     if (T->cache[0]) return shift_cache(T);
 NEXT_CHUNK:
@@ -933,7 +936,7 @@ NEXT_CHUNK:
       /* end and a are odd; sieve[k] contains the a + 8*2k + (0,2,...,14).
        * The largest k is (end-a) >> 4 */
       T->pos = 0;
-      T->maxpos = (T->end - T->a) >> 4;
+      T->maxpos = (T->end - T->a) >> 4; /* >= 1 */
       sieve_block(T->a, T->end, T->maxpos, T->sieve);
       goto NEXT_CHUNK;
     }
